//
// Smile Motor Decoder Easy
// Copyright(C)'2018- ayano / twaydcc / Desktop Station / MECY
//

#include <Arduino.h>
#include <string.h>
#include <avr/pgmspace.h> 
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below
#include "motor_func.h"
#include "motor_ctrl.h"
#include "NmraDcc_diet.h"


/*************************************************
 * 
 * Declaration of Symbols
 * 
 *************************************************/

//各種設定、宣言

#define MAN_VER_NUMBER  10   /* Release Ver 0.1 in CV07 */
#define DECODER_ADDRESS 3

#ifdef _DSFOX
 #define PIN_FHEAD   A3
 #define PIN_RHEAD   A4
 #define RUN_LED      5
#else
 // for Smile Motor Bipolar
 #define PIN_FHEAD    0
 #define PIN_RHEAD    3
#endif

//#define _ROOMLIGHT

#define CV_VSTART     2
#define CV_ACCRATIO   3
#define CV_DECCRATIO  4
#define CV_VMAX       5
#define CV_CONSIST   19
#define CV_BRIGHT    33
#define CV_F0CTRL    34

#define CV_ROOMBRIGHT 35

#define CV_BEMFcoefficient  54
#define CV_PI_P 55
#define CV_PI_I 56
#define CV_BEMFCUTOFF 57

/*************************************************
 * 
 * Declaration of Classes
 * 
 *************************************************/

NmraDcc	 Dcc;
DCC_MSG	 Packet;

/*************************************************
 * 
 * Declaration of Variables
 * 
 *************************************************/


//Task Schedule
unsigned long gPreviousL1 = 0;
//unsigned long gPreviousL2 = 0; // VVVF control
unsigned long gPreviousL3 = 0;
//unsigned long gPreviousL5 = 0; // gIntervalRail

uint8_t gSvcMode = 0;

//モータ制御関連の変数
uint16_t gSpeedCmd = 0;
uint8_t gPwmDir = 128;  // FWD
uint8_t gPwmDirv = 128;

int MOTOR_PWMRef = 0;

//ファンクションの変数
uint8_t gFuncBits[2] = {0};

// CV
uint8_t gCV1_SAddr = 3;
uint16_t gCVx_LAddr = 3;
uint8_t gCV19_CAddr = 0;
uint8_t gConsistDir = 0;

uint8_t gCV5_VMAX = 254;

uint8_t gCV29_Config = 2;
uint8_t gCV29_LAddrEn = 0;
uint8_t gCV33_Bright = 0; // 停車中の減光
uint8_t gCV34_F0Ctrl = 1;
#ifdef _ROOMLIGHT
uint8_t gCV35_RoomBright = 64;
#endif

struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};

CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_29_CONFIG, 2},	 //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
  {CV_VSTART, 2},
  {CV_ACCRATIO, 8},
  {CV_DECCRATIO, 8},
  {CV_VMAX, 255},
  {CV_CONSIST, 0},          // consist address, added by twaydcc 2018/09/30
  {CV_BRIGHT, 3},           // F0 brightness 0:100%,1:75%,2:50%,3:25%
  {CV_F0CTRL, 0},            // F0 ctrl. 0:normal, 1:fwr, 2:non, 3:rev
  {CV_ROOMBRIGHT, 64},
  {CV_BEMFcoefficient, 64}, // 0-127 step / Reajust by MECY 2017/04/16
  {CV_PI_P, 18},
  {CV_PI_I, 96},
  {CV_BEMFCUTOFF,0},        // 0でBEMF Off 1でBEMF On
};

/*************************************************
 * 
 * Declaration of Functions
 * 
 *************************************************/

void(* resetFunc) (void) = 0;  //declare reset function at address 0

uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);

#ifdef _ROOMLIGHT
void ControlRoomLight(void);
#endif
void ControlHeadLight(void);

boolean chkMyAddr(uint16_t inAddr);
boolean chkConsistAddr(uint16_t inAddr);

/*************************************************
 * 
 * Functions
 * 
 *************************************************/
void notifyDccReset(uint8_t hardReset)
{
	gSpeedCmd = 0;
	gFuncBits[0] = 0; gFuncBits[1] = 1;
	
	#ifdef _DSFOX
	Serial.println("DccReset");
	#endif
}

void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

	resetCVToDefault();	 
#ifdef _DSFOX
	Serial.println("Resetting...");
#endif
	delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying
	resetFunc();
}


void resetCVToDefault()
{
	//CVをデフォルトにリセット
#ifdef _DSFOX
	Serial.println("CVs being reset to factory defaults");
#endif
	
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		EEPROM.write( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
}

void notifyServiceMode(uint8_t svc)
{
   gSvcMode = svc;
   
#ifdef _DSFOX
   Serial.print("notifyServiceMode:"); 
   Serial.println(gSvcMode);
#endif
}

uint8_t notifyCVWrite( uint16_t CV, uint8_t Value)
{
   uint16_t aAddr = Dcc.getAddrPacket();
   
   boolean bUpdate = false;
   
   #ifdef _DSFOX
   Serial.print("notifyCVWrite:");
   Serial.print("SvcMode");
   Serial.print(gSvcMode);
   Serial.print(",Addr");
   Serial.print(aAddr);
   Serial.print(",CV");
   Serial.println(CV);
   #endif
   
   if (gSvcMode == 0)
   {
     if (!chkMyAddr(aAddr) && !chkConsistAddr(aAddr)) {
       #ifdef _DSFOX
       Serial.print("out of Address");
       Serial.println(aAddr);
       #endif
       return 0;
     }
   }
  
   //CV変更前のメッセージ
#ifdef _DSFOX
   Serial.print(" Changed to ");
   Serial.println(Value, DEC);
#endif
  
   switch(CV) {
     case CV_MULTIFUNCTION_PRIMARY_ADDRESS:
       gCV1_SAddr = Value;
       bUpdate = true;
       break;
     case CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB:
       gCVx_LAddr = ((Value & 0b00111111) << 8);
       bUpdate = true;
       break;
     case CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB:
       gCVx_LAddr += Value;
       bUpdate = true;
       break;
       
     case CV_VSTART:
     case CV_ACCRATIO:
     case CV_DECCRATIO:
     case CV_VMAX:
       MOTOR_SetCV(CV, Value);
       bUpdate = true;
       break;
       
     case CV_MANUFACTURER_ID:
       if (Value == 8) {
         resetCVToDefault();
       }
       break;
       
     case CV_29_CONFIG:
       MOTOR_SetCV(CV, Value);
       MOTOR_SetDir(gPwmDir);
       bUpdate = true;
       break;
       
     case CV_CONSIST:
       gCV19_CAddr = Value;
       bUpdate = true;
       break;
     case CV_BRIGHT:
       gCV33_Bright = Value;
       bUpdate = true;
       break;
     case CV_F0CTRL:
       gCV34_F0Ctrl = Value;
       bUpdate = true;
       break;
       
     #ifdef _ROOMLIGHT
     case CV_ROOMBRIGHT:
       gCV35_RoomBright = Value;
       bUpdate = true;
       break;
     #endif
       
     case CV_BEMFcoefficient:
     case CV_PI_P:
     case CV_PI_I:
     case CV_BEMFCUTOFF:
       MOTOR_SetCV(CV, Value);
       bUpdate = true;
       break;
   }
   
   if (bUpdate && (Value != EEPROM.read(CV)))
   {
     EEPROM.write( CV, Value );
     
     #ifdef _DSFOX
     Serial.print("write CV");
     Serial.println(CV);
     #endif
   }
   else
   {
     #ifdef _DSFOX
     Serial.print("not write CV");
     Serial.println(CV);
     #endif
   }
   
   return Value;
}

void notifyCVChange( uint16_t CV, uint8_t Value)
{
   //CVが変更されたときのメッセージ
#ifdef _DSFOX
   Serial.print("notifyCVChange:");
   Serial.print(gSvcMode);
   Serial.print(",CV");
   Serial.print(CV); 
   Serial.print(" Changed to"); 
   Serial.println(Value, DEC);
#endif
}

void notifyCVAck(void)
{
  if (gSvcMode != 0)
  {
	MOTOR_Ack();
  }
  
#ifdef _DSFOX
  Serial.print("notifyCVAck:");
  Serial.println(gSvcMode);
#endif
}


void setup()
{
	//シリアル通信開始
	#ifdef _DSFOX
	Serial.begin(115200);
	#endif
	
	//ファンクションの割り当てピン初期化
	#ifdef _DSFOX
	pinMode(RUN_LED, OUTPUT);
	digitalWrite(RUN_LED, 0);
	#endif
	
	pinMode(PIN_FHEAD, OUTPUT);
	digitalWrite(PIN_FHEAD, 0);
	
	pinMode(PIN_RHEAD, OUTPUT);
	digitalWrite(PIN_RHEAD, 0);
	
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
	{
		//if eeprom has 0xFF then assume it needs to be programmed
		#ifdef _DSFOX
		Serial.println("CV Defaulting due to blank eeprom");
		#endif
		
		notifyCVResetFactoryDefault();
		
	} else {
		#ifdef _DSFOX
		Serial.println("CV Not Defaulting");
		#endif
	}
	
	// Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
	Dcc.pin(0, 2, 0);
	
	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( MAN_ID_DIY, MAN_VER_NUMBER, FLAGS_OUTPUT_ADDRESS_MODE, 0 );
	
	//Reset task
	gPreviousL1 = millis();
	gPreviousL3 = millis();
	
	//Init CVs
	gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
	gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) & 0b00111111) << 8;
	gCVx_LAddr += Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
	gCV5_VMAX = Dcc.getCV(CV_VMAX);
	gCV19_CAddr = Dcc.getCV(CV_CONSIST) & 0b01111111;
	gCV29_Config = Dcc.getCV( CV_29_CONFIG ) ;
	gCV29_LAddrEn = ((gCV29_Config & 0x20) > 0) ? 1 : 0;
	
	gCV33_Bright = Dcc.getCV(CV_BRIGHT);
	gCV34_F0Ctrl = Dcc.getCV(CV_F0CTRL);
	#ifdef _ROOMLIGHT
	gCV35_RoomBright = Dcc.getCV(CV_ROOMBRIGHT);
	#endif

	MOTOR_Init();
	MOTOR_SetCV(2, Dcc.getCV(CV_VSTART));
	MOTOR_SetCV(3, Dcc.getCV(CV_ACCRATIO));
	MOTOR_SetCV(4, Dcc.getCV(CV_DECCRATIO));
	MOTOR_SetCV(5, gCV5_VMAX);
	MOTOR_SetCV(29, gCV29_Config);
	MOTOR_SetCV(54, Dcc.getCV(CV_BEMFcoefficient));	//Nagoden Adjust1 2016/06/02
	MOTOR_SetCV(55, Dcc.getCV(CV_PI_P));			// MECY Adjust1 2016/04/09
	MOTOR_SetCV(56, Dcc.getCV(CV_PI_I));			// MECY Adjust1 2016/04/09
	MOTOR_SetCV(57, Dcc.getCV(CV_BEMFCUTOFF));

	#ifdef _DSFOX
	Serial.print("CV1(ShortAddr): ");
	Serial.println(gCV1_SAddr);
	Serial.print("CV17/18(LongAddr): ");
	Serial.println(gCVx_LAddr);
	Serial.print("CV19: ");
	Serial.println(gCV19_CAddr);
	Serial.print("CV29: ");
	Serial.println(gCV29_Config);
	
	Serial.print("CV4: ");
	Serial.println(Dcc.getCV(CV_DECCRATIO));
	Serial.print("CV5: ");
	Serial.println(gCV5_VMAX);
	Serial.print("CV33: ");
	Serial.println(gCV33_Bright);
	Serial.print("CV34: ");
	Serial.println(gCV34_F0Ctrl);
	#ifdef _ROOMLIGHT
	Serial.print("CV35: ");
	Serial.println(gCV35_RoomBright);
	#endif
	
	Serial.println("Ready");
	#endif
}

void loop()
{
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();

	if ((millis() - gPreviousL1) >= 1)
	{
		ControlHeadLight();
		
		gPreviousL1 = millis();
	}

#ifdef _BEMF
  if ((millis() - gPreviousL1) >= 10)
	{
		//Speed detection
		MOTOR_Sensor();
		
		//Reset task
		gPreviousL1 = millis();

	}
#endif
	
#ifdef _ROOMLIGHT
	if ((millis() - gPreviousL3) >= 250)
	{	
		ControlRoomLight();
#else
	if ((millis() - gPreviousL3) >= 100)
	{
		//Motor drive control
		MOTOR_PWMRef = MOTOR_Main(gSpeedCmd, &gPwmDirv);
#endif
		//Reset task
		gPreviousL3 = millis();
	}
}

//固有アドレスチェック
boolean chkMyAddr(uint16_t inAddr)
{
	boolean bMyAddr = false;
	
	// 基本アドレス
	if (gCV29_LAddrEn == 0)
	{
		if (inAddr == gCV1_SAddr) bMyAddr = true;
	}
	// 拡張アドレス
	else
	{
		if (inAddr == gCVx_LAddr) bMyAddr = true;
	}
	
	return bMyAddr;
}

//統括アドレスチェック
boolean chkConsistAddr(uint16_t inAddr)
{
	return (inAddr == gCV19_CAddr);
}

//統括アドレス設定
void notifyDccConsist( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Cmd, uint8_t ConsistAddr)
{
	if (!chkMyAddr(Addr) && !chkConsistAddr(Addr)) return;
	
	gCV19_CAddr = ConsistAddr;
	gConsistDir = Cmd << 7;
	
	// call writeCV() -> notifyCVWrite() in Dcc
	//result = Dcc.setCV(CV_CONSIST, gConsistDir | gCV19_CAddr);
	EEPROM.write( CV_CONSIST, gConsistDir | gCV19_CAddr);
	
	#ifdef _DSFOX
	Serial.print("notifyConsist:");
	Serial.print(Addr);
	Serial.print(",");
	Serial.println(gCV19_CAddr);
	#endif
}

//DCC速度信号の受信によるイベント
void notifyDccSpeed( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Speed, DCC_DIRECTION Dir, DCC_SPEED_STEPS SpeedSteps )
{
	if (gCV19_CAddr != 0) {
		if (!chkConsistAddr(Addr)) return;
	} else {
		if (!chkMyAddr(Addr)) return;
	}
	
	uint16_t aSpeedRef = 0;
	
	//速度値の正規化(255を100%とする処理)
	if (Speed >= 1)
	{
		aSpeedRef = ((Speed - 1) * 255) / SpeedSteps;
	}
	else
	{
		//緊急停止信号受信時の処理 //Nagoden comment 2016/06/11
		aSpeedRef = 0;
		
		#ifdef _DSFOX
		Serial.println("***** Emagency STOP **** ");
		#endif
	}

	gSpeedCmd = aSpeedRef * gCV5_VMAX / 255; 
	
	if (Dir == DCC_DIR_FWD) {
		gPwmDir = 128;
	} else {
		gPwmDir = 0;
	}
	MOTOR_SetDir(gPwmDir);
}

//ファンクション信号受信のイベント
void notifyDccFunc( uint16_t Addr, DCC_ADDR_TYPE AddrType, FN_GROUP FuncGrp, uint8_t FuncState)
{
	boolean bMyAddr = false;
	static uint8_t aFuncBits0 = 0;
	static uint8_t aFuncBits1 = 0;
	
	switch(FuncGrp)
	{
	  case FN_0_4:
		// Function F00
		if (gCV19_CAddr == 0) {
			bMyAddr = chkMyAddr(Addr);
			if(gFuncBits[0] != aFuncBits0)
			{
				#ifdef _DSFOX
				Serial.print("F0:MyAddr ");
				Serial.print(Addr);
				Serial.print(",");
				Serial.println(bMyAddr);
				#endif
			}
			if (!bMyAddr) return;
		} else {
			bMyAddr = chkConsistAddr(Addr);
			if(gFuncBits[0] != aFuncBits0)
			{
				#ifdef _DSFOX
				Serial.print("F0:ConsistAddr ");
				Serial.print(Addr);
				Serial.print(",");
				Serial.println(bMyAddr);
				#endif
				aFuncBits0 = gFuncBits[0];
			}
			if (!bMyAddr) return;
		}
		
		if ((FuncState & FN_BIT_00) == 16)
		{
			if (gFuncBits[0] == 0) gFuncBits[0] = 1;
		}
		else
		{
			if (gFuncBits[0] == 1) gFuncBits[0] = 0;
		}
		
		// Function F01
		bMyAddr = chkMyAddr(Addr);
		if(gFuncBits[1] != aFuncBits1)
		{
			#ifdef _DSFOX
			Serial.print("F1:MyAddr ");
			Serial.print(Addr);
			Serial.print(",");
			Serial.println(bMyAddr);
			#endif
			aFuncBits1 = gFuncBits[1];
		}
		if (!bMyAddr) return;
		
		if ((FuncState & FN_BIT_01) == 1)
		{
			if (gFuncBits[1] == 0) gFuncBits[1] = 1;
		}
		else
		{
			if (gFuncBits[1] == 1) gFuncBits[1] = 0;
		}
		
		#ifdef _DSFOX
		if (gFuncBits[1] == 0)
		{
			exec_function(0, RUN_LED, gFuncBits[0]);
		}
		else if (gFuncBits[0] == 0)
		{
			exec_function(0, RUN_LED, gFuncBits[1]);
		}
		#endif
		
		break;
	}
}

void exec_function (int function, int pin, int FuncState)
{
	digitalWrite (pin, FuncState);
}

#ifdef _ROOMLIGHT
void ControlRoomLight(void)
{
 #ifdef _DSFOX
	digitalWrite(RUN_LED, gFuncBits[1]);
 #endif
	
	pwmWrite(motor_pwmb, 0);
	pwmWrite(motor_pwma, gCV35_RoomBright);
}
#endif

// 1ms interval Task
void ControlHeadLight(void)
{
	static uint8_t aBrightCount = 0;
	uint8_t aHeadOn = 0;
	uint8_t aCV29Dir = MOTOR_GetDirection();
	
	if (gFuncBits[0] == 0)
	{
		F0Control(PIN_FHEAD, aBrightCount, 0);
		F0Control(PIN_RHEAD, aBrightCount, 0);
		return;
	}
	
	aBrightCount++;
	if (aBrightCount > 3)
	{
		aBrightCount = 0;
	}
	
	
	// 前進
	if (gPwmDirv == 128)
	{
		// 後退方向の消灯
		digitalWrite(PIN_RHEAD, 0);
		
		// 進行方向の点灯
		aHeadOn = 1;
		
		// 後続車両は点灯しない
		if (gCV34_F0Ctrl > 1) aHeadOn = 0;
		
		F0Control(PIN_FHEAD, aBrightCount, aHeadOn);
	}
	// 後退
	else
	{
		// 進行方向の消灯
		digitalWrite(PIN_FHEAD, 0);
		
		// 後退方向の点灯
		aHeadOn = 1;
		
		if (gCV34_F0Ctrl == 0)
		{
			// 統括アドレスは点灯しない
			if (gCV19_CAddr != 0)  aHeadOn = 0;
		}
		else if (gCV34_F0Ctrl < 3)
		{
			// 最後尾車両以外は点灯しない
			aHeadOn = 0;
		}
		
		F0Control(PIN_RHEAD, aBrightCount, aHeadOn);
	}
}

void F0Control(uint8_t inPin, uint8_t inCount, uint8_t inOn)
{
	static uint8_t OldOn = 0;
	
	#ifdef _DSFOX
	if (inOn != OldOn)
	{
		Serial.print("F0Control:");
		Serial.print(inPin);
		Serial.print(",");
		Serial.println(inOn);
		
		OldOn = inOn;
	}
	#endif
	
	if (inOn == 0)
	{
		digitalWrite(inPin, 0); // Off
		return;
	}
	
	// 停車中の減光
	if (MOTOR_PWMRef == 0)
	{
		if (inCount < gCV33_Bright)
		{
			digitalWrite(inPin, 0); // Off
		}
		else
		{
			digitalWrite(inPin, 1); // On
		}
	}
	// 進行中
	else
	{
		digitalWrite(inPin, 1); // On
	}
}
