﻿// Motor Control for DS-DCC decode
// By yaasan
//
// MP3V5_DL109T by nagoden

#include <Arduino.h>
#include <string.h>
#include "motor_ctrl.h"
#include "motor_func.h"


/*************************************************
 * 
 * Declaration of Symbols
 * 
 *************************************************/

#define COUNT_ZEROMAX 10
#define MP6513

/*************************************************
 * 
 * Declaration of Classes
 * 
 *************************************************/


/*************************************************
 * 
 * Declaration of Variables
 * 
 *************************************************/

//モータ制御関連の変数
long gMotorLPF_buf = 0;
long gBEMFLPF_buf = 0;

long gMotorSpeed = 0;
byte gDirection = 128;
byte gCV29Direction = 0;
int gPWMRef = 0;

#ifdef _BEMF
long gMotorPI_buf = 0;
byte gZeroCount = 0;
#endif

MOTOR_PARAM gParam;

#if defined(__AVR_ATtiny85__)
 volatile uint8_t* Port[] = {&OCR1B, &OCR1A}; // MR, ML
 int Pin[] = {1, 4};
#else
 int Pin[] = {11, 3};
#endif


/*************************************************
 * 
 * Declaration of Functions
 * 
 *************************************************/

void MOTOR_Init();
void MOTOR_SetCV(byte iNo, byte inData);
void MOTOR_Sensor();
void MOTOR_Ack(void);
int MOTOR_Main(int inSpeedCmd, uint8_t *inDirection);

void setStepSpeed(byte inDirection, int inPWMRef);


/*************************************************
 * 
 * Functions
 * 
 *************************************************/

void pwmWrite(int motor, int intensity)
{
#if defined(__AVR_ATtiny85__)
  *Port[motor] = intensity;
#else
  analogWrite(motor, intensity);
#endif
}

void MOTOR_Init()
{
	gParam.mStartVoltage = 12;
	gParam.mAccRatio = 1;
	gParam.mDecRatio = 1;
	#ifdef _BEMF
	gParam.mBEMFcoefficient = 64;
	gParam.mPI_P = 18;
	gParam.mPI_I = 96;
	gParam.mBEMFcutoff = 32;
	#endif
	
	//PWM キャリア周期
#if defined(__AVR_ATtiny85__)
	//TCCR1 = 1<<PWM1A | 2<<COM1A0 | 1<<CS10;  //31kHz at 8MHz
	TCCR1 = 1<<PWM1A | 2<<COM1A0 | 1<<CS11;  //31kHz at 16MHz
	GTCCR = 1<<PWM1B | 2<<COM1B0;
#else // DSFox
	TCCR2B &= B11111000;
	TCCR2B |= B00000001;
#endif
	
	//PWM出力ピンを出力にセット
	pinMode(Pin[0], OUTPUT);
	pinMode(Pin[1], OUTPUT);
}

void MOTOR_SetCV(byte iNo, byte inData)
{
	switch(iNo)
	{
	  case 2:
		gParam.mStartVoltage = inData;
		break;
		
	  case 3:
		if (inData > 16)
		{
			//gParam.mAccRatio = 255;
			//gParam.mAccRatio = 16;                    // Adjust CV3 Para set by MECY 2017/3/26
			//gParam.mAccRatio = 160;                   // Readjust CV3 Para set by MECY 2017/4/11
			gParam.mAccRatio = 0;                     // Inverted by twaydcc 2018/09/26
		}
		else
		{
			//gParam.mAccRatio = inData * 16;
			//gParam.mAccRatio = inData ;               // Adjust CV3 Para set by MECY 2017/3/26
			//gParam.mAccRatio = inData * 10 ;          // Readjust CV3 Para set by MECY 2017/4/11
			gParam.mAccRatio = (16 - inData) * 10;    // Inverted by twaydcc 2018/09/26
		}
		break;
		
	  case 4:
		if (inData > 16)
		{
			//gParam.mDecRatio = 255;
			//gParam.mDecRatio = 16;                    // Adjust CV4 Para set by MECY 2017/3/26
			//gParam.mDecRatio = 160;                   // Readjust CV4 Para set by MECY 2017/4/11
			gParam.mDecRatio = 0;                     // Inverted by twaydcc 2018/09/26
		}
		else
		{
			//gParam.mDecRatio = inData * 16;
			//gParam.mDecRatio = inData ;               // Adjust CV4 Para set by MECY 2017/3/26
			//gParam.mDecRatio = inData * 10;           // Readjust CV4 Para set by MECY 2017/4/11
			gParam.mDecRatio = (16 - inData) * 10;    // Inverted by twaydcc 2018/09/26
		}
		break;
		
	  case 5:
		gParam.mMaxVoltage = inData;
		break;
		
	  case 29:
		//cv29 Direction Check
		if ( (inData & 0x01) > 0) {
			gCV29Direction = 1;//REVをFWDにする
		} else {
			gCV29Direction = 0;//FWDをFWDにする
		}
		break;
	#ifdef _BEMF
	  case 54:
		gParam.mBEMFcoefficient = inData;
		break;
	  case 55:
		gParam.mPI_P = inData;
		break;
	  case 56:
		gParam.mPI_I = inData;
		break;
	  case 57:
		gParam.mBEMFcutoff = inData;
		/*** 利用停止 MECY 2017/8/6
		if(0 < inData < 20)
		{
			gParam.mBEMFcutoff = 20;                  // Adjust CV57 Para set by MECY 2017/4/16
		}
		else
		{
			gParam.mBEMFcutoff = inData;              // Adjust CV57 Para set by MECY 2017/4/16
		}
		***/
		break;
	#endif
	}
}

void MOTOR_Ack(void)
{
	static byte sDir = 0;
	
	if (sDir == 0) {
		pwmWrite(pwm_ML, 0);
		pwmWrite(pwm_MR, 250);
	} else {
		pwmWrite(pwm_MR, 0);
		pwmWrite(pwm_ML, 250);
	}
	
	delay( 6 );
	
	if (sDir == 0) {
		pwmWrite(pwm_MR, 0);
		sDir = 1;
	} else {
		pwmWrite(pwm_ML, 0);
		sDir = 0;
	}
}

/* Buck EMF Task (100Hz) */
void MOTOR_Sensor()
{
	
	//int aBEMF_result, aSpeed_calculated;
	int aBEMF_result = 0 ;           // BEMF enhance by MECY 2017/09/17
	int aSpeed_calculated = 0;       // BEMF enhance by MECY 2017/09/17
	
	//BEMF detection
	aBEMF_result = MOTOR_GetBEMF();
	
	#ifdef _DSFOX
	Serial.println(aBEMF_result, DEC);
	#endif
	
	aSpeed_calculated = aBEMF_result /6 ;    // BEMF enhance by MECY 2017/09/24
	
	if(aSpeed_calculated > 255)
	{
		aSpeed_calculated = 255;
	}
	
	//Speed conversion
	gMotorSpeed = MOTOR_LPF(aSpeed_calculated, 8, &gBEMFLPF_buf);
	
	
}

void setStepSpeed(byte inDirection, int inPWMRef)
{
	if (inPWMRef == 0)
	{
		pwmWrite(pwm_MR, 0);
		pwmWrite(pwm_ML, 0);
		
		return;
	}
	
	if (inDirection == 128) // FWR
	{
		#ifdef MP6513
		pwmWrite(pwm_MR, 255);
		pwmWrite(pwm_ML, 255 - inPWMRef);
		#else
		pwmWrite(pwm_ML, 0);
		pwmWrite(pwm_MR, inPWMRef);
		#endif
	}
	else // REV
	{
		#ifdef MP6513
		pwmWrite(pwm_ML, 255);
		pwmWrite(pwm_MR, 255 - inPWMRef);
		#else
		pwmWrite(pwm_MR, 0);
		pwmWrite(pwm_ML, inPWMRef);
		#endif
	}
}

void MOTOR_SetDir(uint8_t inDir)
{
	if ( gCV29Direction == 0)
	{
		gDirection = inDir;
	}
	else // 進行方向を反転する
	{
		gDirection = ~inDir & 0b10000000;
	}
}

/* Motor control Task (10Hz) */
int MOTOR_Main(int inSpeedCmd, uint8_t *inDirection)
{
  int aSpeedRef = (double)inSpeedCmd * ((double)gParam.mMaxVoltage - (double)gParam.mStartVoltage) / (double)gParam.mMaxVoltage ;
  int aPWMRef = 0;
  long aSpeedRef_offseted = 0;
  
  static int aPrevSpeed = 0;
  static int aPrevSpeedRef = 0;
  static uint8_t aPrevDir = 128;
  
  //8bit LPF
  if (aPrevSpeedRef <= aSpeedRef)
  {
    aSpeedRef = MOTOR_LPF(aSpeedRef, gParam.mAccRatio, &gMotorLPF_buf);
  }
  else
  {
    aSpeedRef = MOTOR_LPF(aSpeedRef, gParam.mDecRatio, &gMotorLPF_buf);
  }
  
#ifdef _BEMF
  //PI Control
  if (gParam.mBEMFcutoff > 0)
  {
    
    //Speed detection
    int aSpeedDet = (gMotorSpeed * (127 - (int)gParam.mBEMFcoefficient)) / 127;     // 127 step value  /Adjust MECY 2016/12/24
    
    if (aSpeedDet < gParam.mBEMFcutoff)
    {
      //ある閾値以下は速度ゼロと見なす
      aSpeedDet = 0;
    }
    else
    {
      //検出速度のオフセットをする
      aSpeedDet = aSpeedDet - gParam.mBEMFcutoff;
    }
    
    //Command zero check
    if (aSpeedDet == 0)
    {
      gZeroCount++;
      
      if (gZeroCount > COUNT_ZEROMAX)
      {
        gZeroCount = 0;
        gMotorPI_buf = 0;
      }
    }
    else
    {
      gZeroCount = 0;
    }
    
    //BEMF時 PI制御
    if((aSpeedRef == 0) && ( aSpeedDet == 0))
    {
      aSpeedRef_offseted = MOTOR_PI(aSpeedRef - aSpeedDet, gParam.mPI_P, gParam.mPI_I, &gMotorPI_buf) ;
    }
    else if((aSpeedRef >= 1) && ( aSpeedDet >= 0))
    {
      aSpeedRef_offseted = MOTOR_PI(aSpeedRef - aSpeedDet, gParam.mPI_P, gParam.mPI_I, &gMotorPI_buf) + gParam.mStartVoltage ; //CV2 response  /MECY ReAdjust 2017/08/27
    }
  }
  else
#endif
  {
    //PI無効
    if (aSpeedRef == 0)
    {
      aSpeedRef_offseted = aSpeedRef;
    }
    else if (aSpeedRef >= 1)
    {
      aSpeedRef_offseted = aSpeedRef + gParam.mStartVoltage ; //2016/06/19 Modify Nagoden 
    }
  }

  //前回速度として保存
  aPrevSpeed = aSpeedRef_offseted ;

  //Limiter (8bit PWM Output)
  aPWMRef = MOTOR_limitSpeed((uint16_t)(aSpeedRef_offseted));
  
  //進行方向更新 (惰行中は更新しない)
  if ((inSpeedCmd > 0) || (aSpeedRef == 0))
  {
    *inDirection = gDirection;
    if (*inDirection != aPrevDir)
    {
      #ifdef _DSFOX
      Serial.print("MOTOR_Main:PrevDir:");
      Serial.print(aPrevDir);
      Serial.print(",Dir=");
      Serial.print(*inDirection);
      Serial.print(",Cmd=");
      Serial.print(inSpeedCmd);
      Serial.print(",Speed=");
      Serial.println(aSpeedRef);
      #endif
      
      aPrevDir = *inDirection;
    }
  }
  else
  {
    #ifdef _DSFOX
    Serial.print("MOTOR_Main:Cmd=");
    Serial.print(inSpeedCmd);
    Serial.print(",Speed=");
    Serial.print(aSpeedRef);
    Serial.print(",Dir=");
    Serial.println(*inDirection);
    #endif
  }
  
  //PWM出力
  setStepSpeed(*inDirection, aPWMRef);
  
  //バックアップ
  //aPrevSpeedRef = inSpeedCmd;
  aPrevSpeedRef = aSpeedRef;
  gPWMRef = aPWMRef;
  
  return gPWMRef;
}

int MOTOR_GetBEMF()
{
  if (gParam.mBEMFcutoff > 0)
  {
	uint32_t aAve = 0;
	int i;
	
	//Stop PWM
	pwmWrite(pwm_ML, 0);
	pwmWrite(pwm_MR, 0);
	
	delayMicroseconds(300);
	
	for( i = 0; i < 8; i++)
	{
		aAve = aAve + analogRead(A0);
		delayMicroseconds(20);
	}
	
	setStepSpeed(gDirection,  gPWMRef);
	
//	return (aAve >> 3);
	return (aAve >> 6);   // BEMF enhance by MECY 2017/09/17
	
  }
}



int MOTOR_limitSpeed(int inSpeed)
{
	uint16_t aSpeedref = 0;                                   //MECY Adjust 2016/06/04
	
	if (inSpeed >= gParam.mMaxVoltage)
	{
		aSpeedref = gParam.mMaxVoltage;                            //MECY Adjust 2016/06/04
	}
	else if (inSpeed <= 0)
	{
		aSpeedref = 0;                                        //MECY Adjust 2016/06/04
	}
	else if (gParam.mBEMFcutoff == 0 )                        //MECY Adjust 2016/06/04
	{
		aSpeedref = inSpeed;
	}
	else if (gParam.mBEMFcutoff > 0 )
	{
		aSpeedref = inSpeed;
	}
	
	return aSpeedref;
}

byte MOTOR_GetDirection()
{
	return gCV29Direction;
}